import random
import json
from pdfme import PDF

abc = 'abcdefghijklmnñopqrstuvwxyzABCDEFGHIJKLMNÑOPQRSTUVWXYZáéíóúÁÉÍÓÚ'

def gen_word():
    return ''.join(random.choice(abc) for _ in range(random.randint(1, 10)))

def gen_text(n):
    return random.choice(['',' ']) + (' '.join(gen_word() for _ in range(n))) + random.choice(['',' '])

def maybe(n=0.5):
    return random.choices([True, False], [n, 1 - n])[0]

def color(min_=0.5, max_=1):
    return [random.uniform(min_, max_) for _ in range(3)]

def gen_rich_text(n, size=7):

    style_ = {
        'b': 1, 'i': 1, 's': random.triangular(size/2, size, size),
        'f': random.choices(['Helvetica', 'Times', 'Courier'], [3, 1, 1])[0],
        'c': color(), 'bg': color(), 'r': random.triangular(-0.4, 0.4), 'u': 1
    }

    obj = {}
    key = '.'

    if maybe():
        obj['style'] = {k:v for k, v in style_.items() if maybe(.1)}
    else:
        style = []
        for k, v in style_.items():
            if maybe():
                if k in ['b', 'i', 'u']:
                    style.append(k)
                elif k == 'r' and v != 0:
                    style.append(k+':'+str(v))
                elif k in ['bg', 'c']:
                    style.append(k+':'+ (' '.join(str(t) for t in v)))
                else:
                    style.append(k+':'+str(v))
        key += ';'.join(style)

    obj[key] = []
    i = 1
    while n > 0:
        words = min(int(n / 3), random.randint(1, 40))
        if words == 0:
            break
        n -= words
        if i % 2 == 0 and words > 1:
            ans = gen_rich_text(words, size)
            v = [v for k,v in ans.items() if k.startswith('.')][0]
            if ans is not None and len(v) > 0:
                obj[key].append(ans)
        else:
            if len(obj[key]) > 0 and isinstance(obj[key][-1], str):
                obj[key][-1] += gen_text(words)
            else:
                obj[key].append(gen_text(words))
        i += 1

    return obj

def gen_content(size, font_size=4, level=1, max_level=3):
    font_size = max(2.5, font_size)
    cols = random.randint(2,3)
    style = {
        'b': 1, 'i': 1, 'u': 1, 'line_height': random.triangular(1, 1.5),
        'f': random.choices(['Helvetica', 'Times', 'Courier'], [3, 1, 1])[0],
        'text_align': random.choices(['j', 'c', 'l', 'r'], [4, 2, 2, 2])[0],
        'r': random.triangular(-0.4, 0.4), 'c': color(), 'bg': color(),
        'indent': random.triangular(0, 20, 0),
        'margin-left': random.triangular(0, 10, 0),
        'margin-right': random.triangular(0, 10, 0),
        'margin-top': random.triangular(0, 10, 0),
        'margin-bottom': random.triangular(0, 10, 0),
    }
    style = {k:v for k, v in style.items() if maybe(.1)}
    if level == 1 or maybe(0.1):
        style['s'] = -1.6 * level + font_size + 6 - cols
    c = []
    obj = {'style': style, 'content': c, 'cols': {"count": cols}}
    n = int(random.triangular(3, (0.15 * (1 - level) + 1) * size))
    l = random.choice([0,1])

    if level == max_level:
        c.append(gen_text(random.randint(50, 300)))
        return obj

    for i in range(n):
        if i%2 == l:
            ans = gen_content(size, font_size, level + 1, max_level)
            c.append(ans)
        else:
            if maybe(0.5):
                c.append(gen_text(random.randint(50, 300)))
            else:
                c.append({'image': 'tests/image_test.jpg', 'style':
                    {'image_place': 'flow' if maybe(0.7) else 'normal'}
                })

    return obj


def gen_table(rows=None, cols=None):
    rows = 10 if rows is None else rows
    cols = int(random.triangular(1, 7, 2)) if cols is None else cols

    obj = {'content': []}
    obj['widths'] = [random.triangular(3, 6) for _ in range(cols)]
    style_ = {
        'cell_fill': color(), 'cell_margin': random.triangular(5, 20, 5),
        'border_width': random.triangular(1, 3), 'border_color': color(0, 0.6),
        'border_style': random.choice(['solid', 'dotted', 'dashed']),
    }
    obj['style'] = {k:v for k, v in style_.items() if maybe(.15)}

    obj['borders'] = [
        {'pos': 'h::2;', 'width': 1.5, 'color': 'green', 'style': 'dotted'},
        {'pos': 'v;1::2', 'width': 2, 'color': 'red', 'style': 'dashed'},
        {'pos': 'h0,1,-1;', 'width': 2.5, 'color': 'blue', 'style': 'solid'},
    ]

    obj['fills'] = [
        {'pos': '::2;::2', 'color': 0.8},
        {'pos': '1::2;::2', 'color': 0.7},
        {'pos': '::2;1::2', 'color': 0.8},
        {'pos': '1::2;1::2', 'color': 0.7},
    ]

    row_spans = {}
    for i in range(rows):
        row = []
        col_spans = 0
        for j in range(cols):
            if col_spans > 0:
                col_spans -= 1
                row.append(None)
            elif row_spans.get(j, {}).get('rows', 0) > 0:
                row_spans[j]['rows'] -= 1
                col_spans = row_spans[j]['cols']
                row.append(None)
            else:
                prob = random.random()
                if prob < 0.3:
                    element = gen_content(1, max_level=1)
                elif prob < 0.7:
                    element = gen_rich_text(random.triangular(2, 200), 5)
                else:
                    element = {'image': 'tests/image_test.jpg'}
                rowspan = max(1, int(random.triangular(1, min(3, rows - i), 1)))
                colspan = max(1, int(random.triangular(1, min(3, cols - j), 1)))
                col_spans = colspan - 1
                element['rowspan'] = rowspan
                element['colspan'] = colspan
                style = {}
                if maybe(0.1): style['cell_fill'] = color()
                if maybe(0.1): style['cell_margin'] = random.triangular(5, 20, 5)
                element['style'] = style
                row_spans[j] = {'rows': rowspan - 1, 'cols': colspan - 1}
                row.append(element)
        obj['content'].append(row)

    return obj



from copy import deepcopy
import re
from .color import PDFColor

page_sizes = {
    'a5': (419.528, 595.276),
    'a4': (595.276, 841.89),
    'a3': (841.89, 1190.551),
    'b5': (498.898, 708.661),
    'b4': (708.661, 1000.63),
    'jis-b5': (515.906, 728.504),
    'jis-b4': (728.504, 1031.812),
    'letter': (612, 792),
    'legal': (612, 1008),
    'ledger': (792, 1224)
}

def subs(string, *args, **kwargs):
    return string.format(*args, **kwargs).encode('latin')

def process_style(style, pdf=None):
    if style is None:
        return {}
    elif isinstance(style, str):
        if pdf is None:
            return {}
        return deepcopy(pdf.formats[style])
    elif isinstance(style, dict):
        return style
    else:
        raise Exception('style must be a str with the name of a style or dict')

def get_page_size(size):
    if isinstance(size, int):
        return (size, size)
    elif isinstance(size, str):
        return page_sizes[size]
    elif isinstance(size, (list, tuple)):
        return tuple(size)
    else:
        raise Exception('Page size must be a two numbers list or tuple, a number'
            '(for a square page) or any of the following strings: {}'.format(
                ', '. join('"{}"'.format(name) for name in page_sizes.keys())
            ))

def parse_margin(margin):
    if isinstance(margin, dict):
        return margin
        
    if isinstance(margin, str):
        margin = re.split(',| ', margin)
        if len(margin) == 1:
            margin = float(margin)
        else:
            margin = [float(x) for x in margin]

    if isinstance(margin, (int, float)):
        margin = [margin] * 4

    if isinstance(margin, (list, tuple)):
        if len(margin) == 0:
            margin = [0] * 4
        elif len(margin) == 1:
            margin = margin * 4
        elif len(margin) == 2:
            margin = margin * 2
        elif len(margin) == 3:
            margin = margin + [margin[1]]
        elif len(margin) > 4:
            margin = margin[0:4]

        return {k: v for k, v in zip(['top', 'right', 'bottom', 'left'], margin)}
    else:
        raise TypeError('margin property must be of type str, int, list or dict')


def parse_style_str(style_str, fonts):
    style = {}
    for attrs_str in style_str.split(';'):
        attrs = attrs_str.split(':')
        if len(attrs) == 0 or attrs == ['']: continue
        elif len(attrs) == 1:
            attr = attrs[0].strip()
            if not attr in ['b', 'i', 'u']:
                raise ValueError('Style elements with no paramter must '
                    'be whether "b" for bold, "i" for italics(Oblique) or '
                    '"u" for underline.')
            style[attr] = True
        elif len(attrs) == 2:
            attr = attrs[0].strip()
            value = attrs[1].strip()
            if attr == "f":
                if value not in fonts:
                    raise ValueError('Style element "f" must have the name '
                        'of a font family already added.')
                
                style['f'] = value
            elif attr == "c":
                style['c'] = PDFColor(value)
            elif attr == "bg":
                style['bg'] = PDFColor(value)
            elif attrs[0] == "s":
                try:
                    v = float(value)
                    if int(v) == v:
                        v = int(v)
                    style['s'] = v
                except:
                    raise ValueError('Style element value for "s" is wrong:'
                        ' {}'.format(value))
            elif attrs[0] == 'r':
                try: style['r'] = float(value)
                except:
                    raise ValueError('Style element value for "r" is wrong:'
                        ' {}'.format(value))
            else:
                raise ValueError('Style elements with arguments must be "f", '
                    '"s", "c", "r"')

        else:
            raise ValueError('Style elements must be "b", "u", "i", "f", '
                    '"s", "c", "r"')

    return style

def default(value, default_):
    return default_ if value is None else value

def get_char_width(char, size, fonts, font_family, font_mode):
    return size * fonts[font_family][font_mode]['widths'][char] / 1000

def get_word_width(word, size, fonts, font_family, font_mode):
    width = 0
    for char in word:
        width += get_char_width(char, size, fonts, font_family, font_mode)
    return width

def create_graphics(graphics):
    last_fill = last_color = last_line_width = last_line_style = None
    stream = ''
    for g in graphics:
        if g['type'] == 'fill':
            if g['color'] != last_fill:
                last_fill = g['color']
                stream += ' ' + str(last_fill)

            stream += ' {} {} {} {} re F'.format(
                round(g['x'], 3), round(g['y'], 3),
                round(g['width'], 3), round(g['height'], 3)
            )
        
        if g['type'] == 'line':
            if g['color'] != last_color:
                last_color = g['color']
                stream += ' ' + str(last_color)

            if g['width'] != last_line_width:
                last_line_width = g['width']
                stream += ' {} w'.format(round(g['width'], 3))

            if g['style'] == 'dashed':
                line_style = ' 0 J [{} {}] 0 d'.format(round(g['width']*3, 3),
                    round(g['width']*1.5, 3))
            elif g['style'] == 'dotted':
                line_style = ' 1 J [0 {}] {} d'.format(round(g['width']*2, 3),
                    round(g['width'], 3)*0.5)
            else:
                line_style = ' 0 J [] 0 d'

            if line_style != last_line_style:
                last_line_style = line_style
                stream += line_style

            stream += ' {} {} m {} {} l S'.format(
                round(g['x1'], 3), round(g['y1'], 3),
                round(g['x2'], 3), round(g['y2'], 3),
            )

    if stream != '':
        stream = ' q' + stream + ' Q'
    
    return stream

def _roman_five(n, one, five):
    return one * n if n < 4 else (one + five if n == 4 else five)

def _roman_ten(n, one, five, ten):
    return _roman_five(n, one, five) if n < 5 else ((five if n < 9 else '')
        + _roman_five(n - 5, one, ten) if n > 5 else five)

def to_roman(n):
    if not (0 < n < 4000):
        raise Exception('0 < n < 4000')
    roman = ''
    n = str(int(n))
    if len(n) > 0:
        roman = _roman_ten(int(n[-1]), 'I', 'V', 'X')
    if len(n) > 1:
        roman = _roman_ten(int(n[-2]), 'X', 'L', 'C') + roman
    if len(n) > 2:
        roman = _roman_ten(int(n[-3]), 'C', 'D', 'M') + roman
    if len(n) > 3:
        roman = _roman_ten(int(n[-4]), 'M', '', '') + roman
    return roman

def get_paragraph_stream(x, y, text_stream, graphics_stream):
    stream = ''
    x, y = round(x, 3), round(y, 3)
    if graphics_stream != '':
        stream += ' q 1 0 0 1 {} {} cm{} Q'.format(x, y, graphics_stream)
    if text_stream != '':
        stream += ' BT 1 0 0 1 {} {} Tm{} ET'.format(x, y, text_stream)
    return stream



class PDFPage:
    def __init__(self, base, width, height, margin_top=0, margin_bottom=0,
        margin_left=0, margin_right=0
    ):
        self.margin_top = margin_top
        self.margin_bottom = margin_bottom
        self.margin_left = margin_left
        self.margin_right = margin_right

        self.width = width
        self.height = height
        self.go_to_beginning()

        self.stream = base.add({'__stream__': {}})
        self.page = base.add({
            'Type': b'/Page', 'Contents': self.stream.id, 'Resources': {}
        })
        self.x_objects = {}
        self.current_id = 0

    @property
    def y(self):
        return self.height - self._y

    @y.setter
    def y(self, value):
        self._y = self.height - value

    def go_to_beginning(self):
        self.content_width = self.width - self.margin_right - self.margin_left
        self.content_height = self.height - self.margin_top - self.margin_bottom

        self.x = self.margin_left
        self._y = self.height - self.margin_top

    def add(self, content):
        if isinstance(content, str):
            content = content.encode('latin')
        current_id = self.current_id
        self.stream['__stream__'][current_id] = content
        self.current_id += 1
        return current_id

    def add_font(self, font_ref, font_obj_id):
        self.page['Resources'].setdefault('Font', {})
        self.page['Resources']['Font'][font_ref] = font_obj_id

    def add_annot(self, obj, rect):
        if not 'Annots' in self.page:
            self.page['Annots'] = []
        _obj = {'Type': b'/Annot', 'Subtype': b'/Link', 'Rect': rect}
        _obj.update(obj)
        self.page['Annots'].append(_obj)

    def add_link(self, uri_id, rect):
        self.add_annot({'A': uri_id}, rect)

    def add_reference(self, dest, rect):
        self.add_annot({'Dest': dest}, rect)

    def add_image(self, image_obj_id, width, height):
        self.page['Resources'].setdefault('XObject', {})
        if not image_obj_id in self.x_objects:
            image_id = 'Im{}'.format(len(self.page['Resources']['XObject']))
            self.page['Resources']['XObject'][image_id] = image_obj_id
            self.x_objects[image_obj_id] = image_id

        self.add(
            ' q {} 0 0 {} {} {} cm /{} Do Q'.format(
                round(width, 3), round(height, 3), round(self.x, 3),
                round(self._y, 3), self.x_objects[image_obj_id]
            )
        )



from io import BytesIO
from uuid import uuid4

from .parser import parse_obj, PDFObject, PDFRef
from .utils import subs

class PDFBase:
    def __init__(self, version='1.5', trailer=None):
        self.version = version
        self.content = []
        if trailer is None:
            self.trailer = {}
        elif not isinstance(trailer, dict):
            raise ValueError('trailer must be a dict')
        else:
            self.trailer = trailer
        self.count = 1

    def add(self, var):
        if not isinstance(var, (dict, list, tuple, set, bytes, bool, int, float, str, PDFObject)):
            raise TypeError('object type not allowed')
        obj = PDFObject(PDFRef(self.count), var)
        self.content.append(obj)
        self.count += 1
        return obj

    def __getitem__(self, i):
        if i == 0: return None
        return self.content[i - 1]

    def __setitem__(self, i, value):
        if i > 0:
            self.content[i] = value

    def __iter__(self):
        for el in [None] + self.content:
            yield el

    def __len__(self):
        return len(self.content)

    def __str__(self):
        return str(self.content)

    def __repr__(self):
        return str(self.content)

    def output(self, buffer):
        header = subs('%PDF-{}\n%%\x129\x129\x129\n', self.version)
        count = len(header)
        buffer.write(header)

        xref = '\nxref\n0 {}\n0000000000 65535 f \n'.format(self.count)

        for i, obj in enumerate(self.content):
            xref += str(count).zfill(10) + ' 00000 n \n'

            obj_bytes = parse_obj(obj)
            bytes_ = subs('{} 0 obj\n', i + 1) + obj_bytes + '\nendobj\n'.encode('latin')
            count += len(bytes_)
            buffer.write(bytes_)

        self.trailer['Size'] = self.count
        if 'ID' not in self.trailer:
            id_ = lambda: b'<' + str(uuid4()).replace('-', '').encode('latin') + b'>'
            self.trailer['ID'] = [id_(), id_()]
        trailer = parse_obj(self.trailer)

        footer = '\nstartxref\n{}\n%%EOF'.format(count + 1)

        buffer.write((xref + 'trailer\n').encode('latin') + trailer + footer.encode('latin'))


import copy

from .utils import parse_style_str, process_style
from .text import PDFText
from .image import PDFImage


TABLE_PROPERTIES = ('widths', 'borders', 'fills')
PARAGRAPH_PROPERTIES = ('text_align', 'line_height', 'indent',
    'list_text', 'list_style', 'list_indent')

class PDFContent:
    def __init__(self, content, fonts, width, height, x=0, y=0, pdf=None):
        if not isinstance(content, dict):
            raise Exception('content must be a dict')

        self.content = content
        self.finished = False
        self.pdf_content_part = None
        self.setup(x, y, width, height)
        self.fonts = fonts
        self.current_height = 0
        self.pdf = pdf

    def setup(self, x=None, y=None, width=None, height=None):
        if x is not None:
            self.x = x
        if y is not None:
            self.min_y = y
        if width is not None:
            self.width = width
        if height is not None:
            self.height = height

        self.max_y = self.min_y - self.height

    def run(self, x=None, y=None, width=None, height=None):
        self.setup(x, y, width, height)
        self.fills = []
        self.lines = []
        self.parts_ = []
        content_part = self.pdf_content_part
        if content_part is None:
            self.pdf_content_part = content_part = PDFContentPart(
                self.content, self, self.x, self.width, self.min_y, self.max_y,
                last=True
            )
        else:
            content_part.init(self.x, self.width, self.min_y, self.max_y)

        ret = content_part.run()
        self.current_height = content_part.min_y - content_part.max_y
        if ret == 'continue':
            self.finished = True

class PDFContentPart:
    def __init__(
            self, content, pdf_content, min_x, width, min_y, max_y, parent=None,
            last=False, inherited_style=None
        ):
        '''
        content: list of elements (dict) that will be added to the pdf. There are
        currently 3 types of elements, defined by key 'type' in the element dict:
        - '.*' for paragraph element. It can have a 'style' key for a dict
            specifying any of the following paragraph properties: 'text_align',
            'line_height', 'indent', 'list_text', 'list_indent', 'list_style'.
        - 'image' for image element. The string in the 'image' key should contain the
            path of the image to be added.
        - 'content' for content element. The list in the key 'content' can contain any of the
            other types of elements. It can have a 'style' key for a dict
            specifying any properties that can be passed down to its children.

        All of the elements can have a 'style' key holding a dict with any of the
        following keys to change the margin of the elements (or the children, in
        the case of 'content' type): 'margin_left', 'margin_top', 'margin_right',
        'margin_bottom',
        '''

        self.p = pdf_content
        self.parent = parent
        self.is_root = parent is None
        self.last = last

        self.style = {'margin_bottom': 5}
        inherited_style = {} if inherited_style is None else inherited_style
        self.style.update(inherited_style)
        self.style.update(process_style(content.get('style'), self.p.pdf))

        self.column_info = content.get('cols', {'count': 1, 'gap': 10})

        if not isinstance(self.column_info, dict):
            raise TypeError(
                'self.column_info must be a dict:'.format(self.column_info)
            )
        self.elements = content.get('content', [])

        self.section_element_index = 0  # index when the last section jump occured
        self.section_delayed = []  # delayed elements when the last section jump occured
        self.children_indexes = []  # the last state of this element
        self.other_children_indexes = None

        self.init(min_x, width, min_y, max_y)

    def init(self, min_x, width, min_y, max_y):
        self.min_x = min_x
        self.min_y = min_y
        self.go_to_beginning()

        self.full_width = width
        self.max_y = max_y
        self.max_height = self.y - self.max_y
        self.last_bottom = 0

        self.cols_n = self.column_info.get('count', 1)
        self.cols_gap = self.column_info.get('gap', max(width / 25, 7))
        cols_spaces = self.cols_gap * (self.cols_n - 1)
        self.col_width = (width - cols_spaces) / self.cols_n

        self.element_index = self.section_element_index # current index
        self.delayed = copy.deepcopy(self.section_delayed) # current delayed elements
        self.will_reset = False
        self.resetting = False
        self.parts_index = len(self.p.parts_)

        self.minim_diff_last = None
        self.minim_diff = None
        self.minim_forward = None

    def add_delayed(self):
        '''Function to add delayed elements to pdf.

        This function will try to add the delayed elements to the pdf, and it
        can return any of these strings:

        - 'continue' means caller could add all the delayed elements
        - 'break' means a parent element is reseting, and this instance must
          stop
        - 'next' means we need to move to the next section.
        '''
        n = 0
        while n < len(self.delayed):
            ret = self.process(copy.deepcopy(self.delayed[n]), False)
            if ret in ['interrupt', 'break', 'partial_next']:
                return ret

            if ret.get('delayed'):
                self.delayed[n] = copy.deepcopy(ret['delayed'])
            else:
                self.delayed.pop(n)

            if ret.get('next', False):
                return 'next'

            if ret.get('image_flow', False):
                n += 1

        if (
            len(self.delayed) > 0 and
            self.element_index >= len(self.elements) - 1
        ):
            return 'next'

        return 'continue'

    def add_elements(self):
        '''Function to add elements to pdf.

        This function will try to add the elements to the pdf, and it can
        return any of these strings:

        - 'continue' means caller could add all the elements
        - 'break' means a parent element is reseting, and this instance must
          stop
        - 'next' means we need to move to the next section
        '''
        len_elems = len(self.elements) - 1
        while self.element_index <= len_elems:
            last = self.element_index == len_elems
            ret = self.process(self.elements[self.element_index], last=last)
            if ret in ['interrupt', 'break', 'partial_next']:
                return ret

            self.element_index += 1
            if ret.get('delayed'):
                self.delayed.append(ret['delayed'])

            if ret.get('next', False):
                return 'next'

        return 'continue'

    def is_element_resetting(self):
        if self.will_reset or self.resetting:
            continue_reset = self.reset()
            return 'retry' if continue_reset else 'continue'
        else:
            return 'break'

    def process_add_ans(self, ans):
        if ans == 'interrupt':
            return ans
        elif ans == 'partial_next':
            if self.other_children_indexes is None:
                return ans
            else:
                return 'retry'
        elif ans == 'break':
            return self.is_element_resetting()
        elif ans == 'next':
            next_section_ret = self.next_section()
            if next_section_ret == 'break':
                return self.is_element_resetting()
            else:
                return next_section_ret

    def run(self):
        while True:
            action = self.process_add_ans(self.add_delayed())
            if action == 'retry':
                continue
            elif action in ['interrupt', 'break', 'partial_next']:
                return action

            action = self.process_add_ans(self.add_elements())
            if action == 'retry':
                continue
            elif action in ['interrupt', 'break', 'partial_next']:
                return action

            if len(self.delayed) > 0:
                continue

            if not self.resetting and self.cols_n > 1:
                if self.last_child_of_resetting():
                    break
                self.start_resetting()
                if self.will_reset:
                    self.reset()
                else:
                    return 'break'
            else:
                self.minim_forward = False
                if not self.reset():
                    break

        return 'continue'

    def last_child_of_resetting(self):
        parent = self.parent
        if parent:
            if self.last:
                if parent.resetting:
                    parent.mimim_forward = False
                    return True
                else:
                # elif len(parent.delayed) == 0:
                    return parent.last_child_of_resetting()
        return False

    def start_resetting(self):
        parent = self.parent
        if parent and self.last  and parent.cols_n > 1:
            # and len(parent.delayed) == 0
            parent.start_resetting()
            return

        self.will_reset = True

    def reset(self):
        if self.minim_diff_last and self.minim_diff_last - self.minim_diff < 1:
            if self.minim_forward:
                self.minim_diff *= 2
            else:
                return False

        self.will_reset = False
        self.p.parts_ = self.p.parts_[:self.parts_index]
        self.go_to_beginning()

        if self.minim_diff is None:
            self.minim_diff = (self.min_y - self.max_y) / 2
            self.max_y += self.minim_diff
        else:
            self.minim_diff_last = self.minim_diff
            self.minim_diff /= 2
            if self.minim_forward:
                self.max_y -= self.minim_diff
            else:
                self.max_y += self.minim_diff

        self.resetting = True

        self.element_index = self.section_element_index
        self.delayed = copy.deepcopy(self.section_delayed)

        return True

    def go_to_beginning(self):
        self.y = self.min_y
        self.x = self.min_x
        self.column = 0
        self.starting = True

    def next_section(self, children_indexes=None):
        """Goes to the new section of the pdf document.

        It can be called by current node or by a child node.
        When called by current node it returns:
        - True if it's safe to continue.
        - False if it should break because a parent is resetting, and this
          child should end its processing.
        When called by child node it returns the new min_x and min_y in a dict or
        None if a parent has said original caller should initiate a chain reaction
        to break all the elements upwards until it get to the parent (the one that
        has the attribute current_element_ended) that should reset.
        """

        if self.column == self.cols_n - 1:
            if self.resetting:
                self.minim_forward = True
                return 'break'
            else:
                self.section_element_index = self.element_index
                self.section_delayed = copy.deepcopy(self.delayed)
                new_index = {
                    'index': self.element_index,
                    'delayed': copy.deepcopy(self.delayed)
                }
                if children_indexes is None:
                    self.children_indexes = []
                    new_children_indexes = [new_index]
                else:
                    self.children_indexes = copy.deepcopy(children_indexes)
                    new_children_indexes = copy.deepcopy(children_indexes)
                    new_children_indexes.append(new_index)

                if self.is_root:
                    return 'interrupt'

                ret = self.parent.next_section(new_children_indexes)
                if ret in ['interrupt', 'break', 'partial_next']:
                    return ret
                self.min_y = ret['min_y']
                self.min_x = ret['min_x']
                self.parts_index = len(self.p.parts_)
                self.go_to_beginning()
                return 'retry' if children_indexes is None else ret
        else:
            self.column += 1
            self.starting = True
            self.y = self.min_y

            if len(self.delayed) > 0 and children_indexes is not None:
                self.other_children_indexes = copy.deepcopy(children_indexes)
                return 'partial_next'

            return 'retry' if children_indexes is None else \
                {'min_x': self.get_min_x(), 'min_y': self.min_y}

    def get_min_x(self):
        return self.min_x + self.column * (self.col_width + self.cols_gap)

    def update_dimensions(self, style):
        s = style
        self.x = self.get_min_x() + s.get('margin_left', 0)
        self.width = self.col_width - \
            s.get('margin_left', 0) - s.get('margin_right', 0)

        if not self.starting:
            self.y -= self.last_bottom

        self.max_height = max(0, self.y - self.max_y)
        self.last_bottom = s.get('margin_bottom', 0)

    def add_top_margin(self, style):
        if self.starting:
            self.starting = False
        else:
            self.y -=  style.get('margin_top', 0)

    def process(self, element, last=False):
        '''Function to add a single element to the pdf

        This function will add an element to the pdf, using the method
        corresponding to the type of the object (text, image, or another content).
        If some parent element is resetting, it will return None. Else,
        a dict with 2 keys:

        - delayed: if not None, it will contain an element that should be added
        by the caller function in the next section, because it didn't fit in
        the current section.

        - next: if True, the caller function should go to the next section of
        the pdf.
        '''
        if not isinstance(element, (dict, str, list, tuple)):
            element = str(element)
        if isinstance(element, (str, list, tuple)):
            element = {'.': element}

        if not isinstance(element, dict):
            raise TypeError(
                'Elements must be of type dict, str, list or tuple:{}'
                .format(element)
            )

        keys = [key for key in element.keys() if key.startswith('.')]

        style = {}
        style.update(self.style)
        if len(keys) > 0:
            style.update(parse_style_str(keys[0][1:], self.p.fonts))
        element_style = process_style(element.get('style'), self.p.pdf)
        style.update(element_style)

        self.update_dimensions(style)

        

        if len(keys) > 0 or 'paragraph' in element:
            return self.process_text(element, style, element_style, keys)
        elif 'image' in element:
            return self.process_image(element, style)
        elif 'table' in element or 'table_delayed' in element:
            return self.process_table(element, style, element_style)
        elif 'content' in element:
            return self.process_child(element, style, last)

    def process_text(self, element, style, element_style, paragraph_keys):
        if 'paragraph' in element:
            pdf_text = element['paragraph']
            pdf_text.setup(
                self.x, self.y, self.width, self.max_height,
                element['last_part'], element['last_word']
            )
            remaining = copy.deepcopy(element)
        else:
            par_style = {
                v: style.get(v) for v in PARAGRAPH_PROPERTIES if v in style
            }
            key = paragraph_keys[0]
            pdf_text = PDFText(
                {key: element[key], 'style': style.copy()},
                self.width, self.max_height, self.x, self.y,
                fonts=self.p.fonts, pdf=self.p.pdf, **par_style
            )
            remaining = {'paragraph': pdf_text, 'style': element_style}

        result = pdf_text.run()
        remaining['last_part'] = pdf_text.last_part
        remaining['last_word'] = pdf_text.last_word
        result['type'] = 'paragraph'
        self.p.parts_.append(result)
        self.y -= pdf_text.current_height

        if pdf_text.current_height > 0:
            self.add_top_margin(style)

        if pdf_text.finished:
            return {'delayed': None, 'next': False}
        else:
            return {'delayed': remaining, 'next': True}

    def process_image(self, element, style):
        ret = {'delayed': None, 'next': False}
        pdf_image = PDFImage(
            element['image'], element.get('extension'),
            element.get('image_name')
        )
        height = self.width * pdf_image.height / pdf_image.width

        if height < self.max_height:
            self.p.parts_.append({
                'pdf_image': pdf_image, 'type': 'image', 'x': self.x,
                'y': self.y - height, 'width': self.width, 'height': height
            })
            self.y -= height
            self.add_top_margin(style)
        else:
            image_place = style.get('image_place', 'flow')
            ret['delayed'] = element
            if image_place == 'normal':
                ret['next'] = True
            elif image_place == 'flow':
                ret['image_flow'] = True
        return ret

    def process_table(self, element, style, element_style):
        if 'table_delayed' in element:
            pdf_table = element['table_delayed']
            pdf_table.setup(self.x, self.y, self.width, self.max_height)
            remaining = element
        else:
            table_props = {
                v: element.get(v) for v in TABLE_PROPERTIES
                if v in element
            }
            pdf_table = PDFTable(
                element['table'], self.p.fonts, self.width, self.max_height,
                self.x, self.y, style=style, pdf=self.p.pdf, **table_props
            )
            remaining = {'table_delayed': pdf_table, 'style': element_style}

        pdf_table.run()
        self.p.parts_.extend(pdf_table.parts_)
        self.p.lines.extend(pdf_table.lines)
        self.p.fills.extend(pdf_table.fills)

        self.y -= pdf_table.current_height
        if pdf_table.current_height > 0:
            self.add_top_margin(style)

        if not pdf_table.finished:
            return {'delayed': remaining, 'next': True}
        else:
            return {'delayed': None, 'next': False}

    def process_child(self, element, style, last):
        pdf_content = PDFContentPart(
            element, self.p, self.get_min_x(), self.col_width, self.y,
            self.max_y, self, last, copy.deepcopy(style)
        )
        down_condition1 = len(self.children_indexes) > 0 and \
            self.element_index == self.section_element_index
        down_condition2 = self.other_children_indexes is not None

        if down_condition1 or down_condition2:
            child = self.children_indexes[-1] if down_condition1 else \
                self.other_children_indexes[-1]
            pdf_content.section_element_index = child['index']
            pdf_content.section_delayed = copy.deepcopy(child['delayed'])
            pdf_content.element_index = child['index']
            pdf_content.delayed = copy.deepcopy(child['delayed'])
            pdf_content.children_indexes = self.children_indexes[:-1] \
                if down_condition1 else self.other_children_indexes[:-1]

            self.other_children_indexes = None

        action = pdf_content.run()

        current_height = pdf_content.min_y - pdf_content.max_y
        self.y -= current_height

        if current_height > 0:
            self.add_top_margin(style)

        if action in ['interrupt', 'break', 'partial_next']:
            return action
        else:
            self.starting = False
            return {'delayed': None, 'next': False}

from .table import PDFTable



from copy import deepcopy
import re

from .encoders import encode_stream
from .utils import subs

class PDFObject:
    def __init__(self, id_=None, obj=None):
        if not isinstance(id_, PDFRef):
            raise TypeError('id_ argument must be of type PDFRef')
        self.id = id_
        self.value = {} if obj is None else obj

    def __getitem__(self, name):
        return self.value[name]

    def __setitem__(self, name, value):
        self.value[name] = value

    def __delitem__(self, name):
        del self.value[name]

    def __contains__(self, name):
        return name in self.value

class PDFRef(int):
    def __new__(cls, id_):
        return int.__new__(cls, id_)
    @property
    def ref(self):
        return subs('{} 0 R', self)

def parse_obj(obj):
    if isinstance(obj, PDFObject):
        return parse_obj(obj.value)
    elif isinstance(obj, PDFRef):
        return obj.ref
    elif isinstance(obj, dict):
        if '__stream__' in obj:
            return parse_stream(obj)
        else:
            return parse_dict(obj)
    elif isinstance(obj, (list, tuple, set)):
        return parse_list(obj)
    elif isinstance(obj, bytes):
        return obj
    elif isinstance(obj, bool):
        return b'true' if obj else b'false'
    elif isinstance(obj, (int, float)):
        return str(obj).encode('latin')
    elif isinstance(obj, str):
        return ('(' + re.sub(r'([()])', r'\\\1', obj) + ')').encode('latin')
    

def parse_dict(obj):
    bytes_ = b'<<'
    for key, value in obj.items():
        bytes_ += b'/' + key.encode('latin')
        ret = parse_obj(value)
        if not ret[0] in [b'/', b'(', b'<']: bytes_ += b' '
        bytes_ += ret

    return bytes_ + b'>>'

def parse_list(obj):
    bytes_ = b'['
    for i, value in enumerate(obj):
        ret = parse_obj(value)
        if not ret[0] in [b'/', b'(', b'<'] and i != 0: bytes_ += b' '
        bytes_ += ret

    return bytes_ + b']'

def parse_stream(obj):
    stream_ = obj.pop('__stream__')
    stream_copy = deepcopy(stream_)
    skip_filter = obj.pop('__skip_filter__', False)

    if isinstance(stream_copy, bytes):
        stream_str = stream_copy
    elif isinstance(stream_copy, dict):
        stream_str = b''.join(stream_copy.values())

    if 'Filter' in obj and not skip_filter:
        stream = encode_stream(stream_str, obj['Filter'])
    else:
        stream = stream_str

    obj['Length'] = len(stream)
    ret = parse_dict(obj) + b'stream\n' + stream + b'\nendstream'
    obj['__stream__'] = stream_
    if skip_filter:
        obj['__skip_filter__'] = True
    return ret



import json
from pathlib import Path

from .utils import gen_rich_text
from pdfme import PDF

def page_rect(pdf):
    pdf.page.add('q 0.9 g {} {} {} {} re F Q'.format(
        pdf.margin['left'], pdf.margin['bottom'],
        pdf.page.content_width, pdf.page.content_height
    ))

def output(pdf, name):
    with open(name, 'wb') as f:
        pdf.output(f)

def add_content(content, text_options, name):
    pdf = PDF()
    pdf.add_page()
    page_rect(pdf)
    pdf_text = pdf._text(
        content, x=pdf.page.margin_left, width=pdf.page.content_width,
        **text_options
    )
    while not pdf_text.finished:
        pdf.add_page()
        page_rect(pdf)
        pdf_text = pdf._text(pdf_text)
    output(pdf, name + '.pdf')


def base(text_options={}, name='test', words=5000):
    input_file = Path(name + '.json')
    if input_file.exists():
        with input_file.open() as f:
            content = json.load(f)
    else:
        content = gen_rich_text(words)
        with input_file.open('w') as f:
            json.dump(content, f, ensure_ascii=False)
    add_content(content, text_options, name)

def test_text_indent():
    base({'indent': 20, 'text_align': 'j'}, 'test_text_indent')

def test_text_line_height():
    base({'line_height': 2}, 'test_text_line_height')

def test_text_left():
    base({}, 'test_text')

def test_text_right():
    base({'text_align': 'r'}, 'test_text_right')

def test_text_center():
    base({'text_align': 'c'}, 'test_text_center')

def test_text_justify():
    base({'text_align': 'j'}, 'test_text_justify')

def test_text_list():
    base({'list_text': '1. '}, 'test_text_list', 500)

def test_text_list_style():
    base({'list_text': chr(183) + ' ', 'list_style': {'f': 'Symbol'}}, 
        'test_text_list_style', 500)

def test_text_list_style_indent():
    base(
        {
            'list_text': chr(183) + ' ', 'list_style': {'f': 'Symbol'},
            'list_indent': 40, 'indent': 20
        },
        'test_text_list_style_indent', 500
    )

def get_content_list(content):
    for key, val in content.items():
        if key.startswith('.'):
            return val

def append_text(content, new):
    for key, val in content.items():
        if key.startswith('.'):
            val.extend(new)
            break

def test_text_ref_label():
    content = gen_rich_text(1)
    append_text(content, [{'ref': 'asdf', '.': 'asd fa sdf asdf asdfa sdfa'}])
    append_text(content, get_content_list(gen_rich_text(500)))
    append_text(content, [{'label': 'asdf', '.': 'ertyer sdfgsd'}])
    append_text(content, get_content_list(gen_rich_text(1000)))
    add_content(content, {}, 'test_text_ref_label')

def test_text_link():
    content = gen_rich_text(500)
    append_text(content, [{'uri': 'www.google.com', '.': 'its me google '*10}])
    append_text(content, get_content_list(gen_rich_text(1000)))
    append_text(content, [{'uri': 'www.google.com', '.': 'its me google '*10}])
    append_text(content, get_content_list(gen_rich_text(500)))
    add_content(content, {}, 'test_text_link')


from copy import deepcopy
import re
from .color import PDFColor
from .utils import parse_style_str, default, process_style, get_paragraph_stream
from .standard_fonts import STANDARD_FONTS

PARAGRAPH_DEFAULTS = {'text_align': 'l', 'line_height': 1.1, 'indent': 0}
TEXT_DEFAULTS = {'f': 'Helvetica', 'c': 0.1, 's': 11, 'r': 0, 'bg': None}


class PDFState:
    def __init__(self, style, fonts=None):

        self.fonts = STANDARD_FONTS if fonts is None else fonts

        self.font_family = style['f']

        f_mode = ''
        if style.get('b', False):
            f_mode += 'b'
        if style.get('i', False):
            f_mode += 'i'
        if f_mode == '':
            f_mode = 'n'
        self.font_mode = 'n' if not f_mode in fonts[style['f']] else f_mode

        self.size = style['s']
        self.color = PDFColor(style['c'])
        self.rise = style.get('r', 0) * self.size

    def compare(self, other):
        ret_value = ''
        if (
            other is None or self.font_family != other.font_family or
            self.font_mode != other.font_mode or self.size != other.size
        ):
            ret_value += ' /{} {} Tf'.format(
                self.fonts[self.font_family][self.font_mode]['ref'],
                round(self.size, 3)
            )
        if other is None or self.color != other.color:
            ret_value += ' ' + str(self.color)
        if other is None or self.rise != other.rise:
            ret_value += ' {} Ts'.format(round(self.rise, 3))

        return ret_value


class PDFTextLinePart:
    def __init__(self, style, fonts=None, ids=None):

        self.fonts = STANDARD_FONTS if fonts is None else fonts

        self.style = style
        self.state = PDFState(style, fonts)
        self.underline = style.get('u', False)
        self.background = PDFColor(style.get('bg'))
        self.ids = [] if id is None else ids
        self.width = 0
        self.words = []

        self.space_width = self.get_char_width(' ')
        self.spaces_width = 0

    def pop_word(self, index=None):
        if len(self.words) > 0:
            word = self.words.pop() if index is None else self.words.pop(index)
            if word == ' ':
                self.spaces_width -= self.space_width
            else:
                self.width -= self.get_word_width(word)
            return word

    def add_word(self, word):
        self.words.append(word)
        if word == ' ':
            self.spaces_width += self.space_width
        else:
            self.width += self.get_word_width(word)

    def current_width(self, factor=1):
        return self.width + self.spaces_width*factor

    def tentative_width(self, word, factor=1):
        word_width = self.space_width * factor if word == ' ' else \
            self.get_word_width(word)
        return self.current_width(factor) + word_width

    def get_char_width(self, char):
        ws = self.fonts[self.state.font_family][self.state.font_mode]['widths']
        return self.state.size * ws[char] / 1000

    def get_word_width(self, word):
        return sum(self.get_char_width(char) for char in word)

class PDFTextLine:
    def __init__(
        self, fonts=None, max_width=0, text_align=None, line_height=None,
        top_margin=0
    ):
        self.fonts = STANDARD_FONTS if fonts is None else fonts
        self.max_width = max_width
        self.line_parts = []

        self.justify_min_factor = 0.7

        self.text_align = default(text_align, PARAGRAPH_DEFAULTS['text_align'])
        self.line_height = default(
            line_height, PARAGRAPH_DEFAULTS['line_height']
        )

        self.top_margin = top_margin
        self.next_line = None
        self.is_last_word_space = True
        self.firstWordAdded = False
        self.started = False

    @property
    def height(self):
        top = 0
        height_ = 0
        for part in self.line_parts:
            if part.state.rise > 0 and part.state.rise > top:
                top = part.state.rise
            if part.state.size > height_:
                height_ = part.state.size

        return height_ + self.top_margin + top

    @property
    def min_width(self):
        ws = self.get_widths()
        return ws[0] + ws[1] * self.factor

    @property
    def factor(self):
        return 1 if self.text_align != 'j' else self.justify_min_factor

    @property
    def bottom(self):
        bottom = 0
        for part in self.line_parts:
            if part.state.rise < 0 and -part.state.rise > bottom:
                bottom = -part.state.rise
        return bottom

    def get_widths(self):
        words_width = 0
        spaces_width = 0
        for part in self.line_parts:
            words_width += part.width
            spaces_width += part.spaces_width
        return words_width, spaces_width

    def add_line_part(self, style=None, ids=None):
        if self.next_line is None:
            self.next_line = PDFTextLine(
                self.fonts, self.max_width, self.text_align, self.line_height
            )

        line_part = PDFTextLinePart(style, self.fonts, ids)
        self.next_line.line_parts.append(line_part)
        return line_part

    def add_accumulated(self):
        if len(self.line_parts):
            for word in self.next_line.line_parts[0].words:
                self.line_parts[-1].add_word(word)
            self.next_line.line_parts = self.next_line.line_parts[1:]

        self.line_parts.extend(self.next_line.line_parts)
        last_part = self.line_parts[-1]
        last_part.add_word(' ')
        self.next_line.line_parts = [
            PDFTextLinePart(last_part.style, self.fonts, last_part.ids)
        ]

    def add_word(self, word):
        if not self.started:
            if word.isspace():
                if self.firstWordAdded:
                    self.started = True
                    self.add_accumulated()
                    return {'status': 'added'}
                else:
                    return {'status': 'ignored'}
            else:
                self.firstWordAdded = True
                self.next_line.line_parts[-1].add_word(word)
                return {'status': 'added'}
        else:
            if word.isspace():
                if self.is_last_word_space:
                    return {'status': 'ignored'}
                else:
                    self.add_accumulated()
                    return {'status': 'added'}
            else:
                self.is_last_word_space = False
                self.next_line.line_parts[-1].add_word(word)
                if (self.min_width + self.next_line.min_width < self.max_width):
                    return {'status': 'preadded'}
                else:
                    if (
                        len(self.line_parts[-1].words) and
                        self.line_parts[-1].words[-1] == ' '
                    ):
                        self.line_parts[-1].pop_word(-1)
                    self.next_line.firstWordAdded = True
                    self.next_line.top_margin = self.bottom
                    self.next_line.next_line = PDFTextLine(
                        self.fonts, self.max_width, self.text_align,
                        self.line_height
                    )
                    line_parts = self.next_line.line_parts
                    self.next_line.next_line.line_parts = line_parts
                    self.next_line.line_parts = []
                    return {
                        'status': 'finished', 'new_line': self.next_line
                    }

class PDFTextBase:
    def __init__(
        self, content, width, height, x=0, y=0, fonts=None, text_align=None,
        line_height=None, indent=0, list_text=None, list_indent=None,
        list_style=None, pdf=None
    ):
        self.fonts = STANDARD_FONTS if fonts is None else fonts
        self.setup(x, y, width, height)
        self.indent = indent
        self.text_align = default(text_align, PARAGRAPH_DEFAULTS['text_align'])
        self.line_height = default(
            line_height, PARAGRAPH_DEFAULTS['line_height']
        )
        self.list_text = list_text
        self.list_indent = list_indent
        self.list_style = list_style
        self.pdf = pdf

        if isinstance(content, str):
            content = [{'style': TEXT_DEFAULTS.copy(), 'text': content}]
        if not isinstance(content, (list, tuple)):
            raise TypeError(
                'content must be of type str, list or tuple: {}'.format(content)
            )

        self.last_part_added = 0
        self.last_part_line = 0
        self.last_part = 0

        self.last_word_added = 0
        self.last_word_line = 0
        self.last_word = 0

        self.content = content
        self.finished = False
        self.is_first_line = True
        self.correct_indent = True
        self.list_setup_done = False

    @property
    def stream(self):
        return get_paragraph_stream(self.x, self.y, self.text, self.graphics)

    @property
    def result(self):
        return dict(
            x=self.x, y=self.y, height=self.current_height, width=self.width,
            text_stream=self.text, graphics_stream=self.graphics,
            used_fonts=self.used_fonts, ids=self.ids,
        )

    def move(self, x, y):
        self.x = x
        self.y = y

    def setup(
        self, x=None, y=None, width=None, height=None, last_part=None,
        last_word=None
    ):
        if x is not None:
            self.x = x
        if y is not None:
            self.y = y
        if width is not None:
            self.width = width
        if height is not None:
            self.height = height
        if last_part is not None:
            self.last_part = last_part
        if last_word is not None:
            self.last_word = last_word

    def init(self):
        self.started = False
        self.lines = []
        self.text = ''
        self.graphics = ''
        self.ids = {}
        self.first_line_added = False

        self.used_fonts = set()
        self.current_line_used_fonts = set()
        self.current_height = 0
        self.lines = []

        line_width = self.width - (self.indent if self.is_first_line else 0)
        self.current_line = PDFTextLine(
            self.fonts, line_width, self.text_align, self.line_height
        )

        self.last_indent = 0
        self.last_state = self.last_factor = self.last_fill = None
        self.last_color = self.last_stroke_width = None

        self.y_ = 0

    def run(
        self, x=None, y=None, width=None, height=None, last_part=None,
        last_word=None
    ):
        self.setup(x, y, width, height, last_part, last_word)
        self.init()
        for part_index in range(self.last_part, len(self.content)):
            part = self.content[part_index]
            if not isinstance(part, dict):
                raise TypeError(
                    'elements in content must be of type dict: {}'
                    .format(part)
                )
            if 'type' in part:
                if part['type'] == 'br':
                    self.last_part_added = part_index + 1
                    self.last_word_added = 0
                    continue_ = self.add_current_line(True)
                    self.current_line = PDFTextLine(
                        self.fonts, self.width, self.text_align,
                        self.line_height
                    )
                    if not continue_:
                        return self.result
            else:
                continue_ = self.add_part(part, part_index)
                if not continue_:
                    return self.result

        continue_ = self.add_current_line(True)
        if continue_:
            self.finished = True
        return self.result

    def add_part(self, part, part_index):
        if 'var' in part and self.pdf:
            part['text'] = str(self.pdf.context.get(part['var'], ''))

        words = part.get('text')
        if not isinstance(words, (str, list, tuple)):
            return 'continue'

        style = TEXT_DEFAULTS.copy()
        style.update(part.get('style', {}))
        new_line_part = self.current_line.add_line_part(
            style=style, ids=part.get('ids')
        )

        if not self.list_setup_done and self.list_text:
            self.list_setup_done = True
            self.setup_list()

        self.current_line_used_fonts.add((
            new_line_part.state.font_family,
            new_line_part.state.font_mode
        ))

        if isinstance(words, str):
            part['text'] = words = [
                ' ' if w.isspace() else w
                for w in re.split('( +)', words)
                if w != ''
            ]
        is_last_part = part_index == len(self.content) - 1

        for word_index in range(self.last_word, len(words)):
            word = words[word_index]
            ans = self.current_line.add_word(word)
            if ans['status'] == 'added':
                self.last_part_added = part_index
                self.last_word_added = word_index + 1
            elif ans['status'] == 'finished':
                continue_ = self.add_current_line(
                    is_last_part and word_index == len(words) - 1
                )
                ans['new_line'].max_width = self.width - (
                    self.list_indent if self.list_text else 0
                )
                ans['new_line'].next_line.max_width = ans['new_line'].max_width
                self.current_line = ans['new_line']
                if not continue_:
                    return False

        self.last_word = 0
        return True

    def add_current_line(self, is_last=False):
        if is_last and self.current_line.next_line is not None:
            self.current_line.line_parts.extend(
                self.current_line.next_line.line_parts
            )

        line_height = self.current_line.height
        if self.first_line_added:
            line_height *= self.line_height
        else:
            self.first_line_added = True

        if line_height + self.current_height > self.height:
            self.last_part = self.last_part_line
            self.last_word = self.last_word_line
            return False
        else:
            self.last_part_line = self.last_part_added
            self.last_word_line = self.last_word_added
            self.current_height += line_height
            self.lines.append(self.current_line)
            self.used_fonts.update(self.current_line_used_fonts)
            self.current_line_used_fonts = set()

            self.add_line_to_stream(self.current_line, is_last)
            self.current_line = None

            return True

    def setup_list(self):
        style = self.current_line.next_line.line_parts[0].style.copy()

        if self.list_style is None:
            self.list_style = {}
        elif isinstance(self.list_style, str):
            self.list_style = process_style(self.list_style, self.pdf)

        if not isinstance(self.list_style, dict):
            raise TypeError(
                'list_style must be a str or a dict. Value: {}'
                .format(self.list_style)
            )

        style.update(self.list_style)
        line_part = PDFTextLinePart(style, self.fonts)

        self.current_line_used_fonts.add(
            (line_part.state.font_family, line_part.state.font_mode)
        )

        if self.list_indent is None:
            self.list_indent = line_part.get_word_width(str(self.list_text))
        elif not isinstance(self.list_indent, (float, int)):
            raise TypeError(
                'list_indent must be int or float. Value: {}'
                .format(self.list_style)
            )

        self.list_state = line_part.state
        self.current_line.max_width -= self.list_indent

    def add_line_to_stream(self, line, is_last=False):
        words_width, spaces_width = line.get_widths()
        x = self.list_indent if self.list_text else 0
        line_height = line.height
        full_line_height = line_height
        ignore_factor = self.text_align != 'j' or is_last or spaces_width == 0
        factor_width = self.width - words_width - x
        adjusted_indent = 0
        if self.text_align in ['r', 'c']:
            indent = self.width - words_width - spaces_width
            if self.text_align == 'c':
                indent /= 2
            x += indent
            adjusted_indent = indent - self.last_indent
            self.last_indent = indent

        if not self.started:
            self.started = True
            if self.is_first_line:
                factor_width -= self.indent
                x += self.indent
                self.is_first_line = False

                first_indent = (
                    adjusted_indent if self.text_align in ['r', 'c']
                    else self.indent
                )

                if self.list_text:
                    first_indent += self.list_indent
                    if self.list_state.size > full_line_height:
                        full_line_height = self.list_state.size
                    self.text += ' 0 -{} Td{} ({})Tj {} 0 Td'.format(
                        round(full_line_height, 3),
                        self.list_state.compare(self.last_state),
                        self.list_text, first_indent
                    )
                else:
                    self.text += ' {} -{} Td'.format(
                        round(first_indent, 3), round(full_line_height, 3)
                    )
            else:
                first_indent = adjusted_indent
                if self.list_text:
                    first_indent += self.list_indent
                self.text += ' {} -{} Td'.format(
                    round(first_indent, 3), round(full_line_height, 3)
                )
        else:
            if self.correct_indent:
                self.correct_indent = False
                adjusted_indent -= self.indent

            full_line_height *= self.line_height

            self.text += ' {} -{} Td'.format(round(adjusted_indent, 3),
                                        round(full_line_height, 3))

        self.y_ -= full_line_height

        factor = 1 if ignore_factor else factor_width / spaces_width

        for part in line.line_parts:
            text = self.clean_words(part.words)
            self.text += self.output_text(part, text, factor)
            part_width = part.current_width(factor)
            part_size = round(part.state.size, 3)

            if text != '' and not text.isspace():
                if part.ids is not None:
                    for id_ in part.ids:
                        id_y = self.y_ + part.state.rise - part_size*0.25
                        self.ids.setdefault(id_, []).append([
                            round(x, 3), round(id_y, 3),
                            round(x + part_width, 3), round(id_y + part_size, 3)
                        ])

                part_graphics = self.output_graphics(
                    part, x, self.y_, part_width
                )
                self.graphics += part_graphics
            x += part_width

    def clean_words(self, words):
        text = ''.join(word for word in words)
        if text != '':
            text = text.replace('\\',r'\\').replace('(','\(').replace(')','\)')
        return text

    def output_text(self, part, text, factor=1):
        stream = part.state.compare(self.last_state)
        self.last_state = part.state

        tw = round(part.space_width * (factor - 1), 3)
        if self.last_factor != tw:
            if tw == 0:
                tw = 0
            stream += ' {} Tw'.format(tw)
            self.last_factor = tw

        if text != '':
            stream += ' ({})Tj'.format(text)
        return stream

    def output_graphics(self, part, x, y, part_width):
        graphics = ''
        if part.background is not None and not part.background.color is None:
            if part.background != self.last_fill:
                self.last_fill = part.background
                graphics += ' ' + str(self.last_fill)

            graphics += ' {} {} {} {} re F'.format(
                round(x, 3),
                round(y + part.state.rise - part.state.size*0.25, 3),
                round(part_width, 3), round(part.state.size*1.2, 3)
            )

        if part.underline:
            color = PDFColor(part.state.color, True)
            stroke_width = part.state.size * 0.07
            y_u = round(y + part.state.rise - stroke_width, 3)

            if color != self.last_color:
                self.last_color = color
                graphics += ' ' + str(self.last_color)

            if stroke_width != self.last_stroke_width:
                self.last_stroke_width = stroke_width
                graphics += ' {} w'.format(round(self.last_stroke_width, 3))

            graphics += ' {} {} m {} {} l S'.format(
                round(x, 3), y_u, round(x + part_width, 3), y_u
            )

        return graphics

class PDFText(PDFTextBase):
    def __init__(
        self, content, width, height, x=0, y=0, fonts=None, text_align=None,
        line_height=None, indent=0, list_text=None, list_indent=None,
        list_style=None, pdf=None
    ):
        self.pdf = pdf
        self.fonts = STANDARD_FONTS if fonts is None else fonts
        self.content = []
        self._recursive_content_parse(content, TEXT_DEFAULTS, [])
        super().__init__(
            self.content, width, height, x, y, fonts, text_align, line_height,
            indent, list_text, list_indent, list_style, pdf
        )

    def _new_text_part(self, style, ids, part_var, last_part=None):
        if last_part is not None and last_part['text'] == '':
            self.content.remove(last_part)
        text_part = {'style': style, 'text': '', 'ids': ids}
        if part_var is not None:
            text_part['var'] = part_var
        self.content.append(text_part)
        return text_part

    def _recursive_content_parse(self, content, parent_style, ids):
        style = deepcopy(parent_style)
        ids = deepcopy(ids)

        if isinstance(content, str):
            content = {'.': [content]}
        elif isinstance(content, (list, tuple)):
            content = {'.': content}

        if not isinstance(content, dict):
            raise TypeError(
                'content must be of type dict, str, list or tuple: {}'
                .format(content)
            )

        elements = []
        for key, value in content.items():
            if key.startswith('.'):
                style.update(parse_style_str(key[1:], self.fonts))
                if isinstance(value, str):
                    value = [value]
                if not isinstance(value, (list, tuple)):
                    raise TypeError(
                        'value of . attr must be of type str, list or tuple: {}'
                        .format(value)
                    )
                elements = value
                break

        style.update(process_style(content.get('style'), self.pdf))
        part_var = content.get('var')
        text_part = self._new_text_part(style, ids, part_var)
        text_part['ids'].extend(content.get('ids', []))

        if part_var is not None:
            elements = ['0']

        label = content.get('label', None)
        if label is not None:
            text_part['ids'].append('$label:' + label)
        ref = content.get('ref', None)
        if ref is not None:
            text_part['ids'].append('$ref:' + ref)
        uri = content.get('uri', None)
        if uri is not None:
            text_part['ids'].append('$uri:' + uri)

        is_last_string = False

        for element in elements:
            if isinstance(element, str):
                if element == '':
                    continue
                lines = element.split('\n')
                if not is_last_string:
                    text_part = self._new_text_part(
                        style, text_part['ids'], part_var, text_part
                    )
                text_part['text'] += lines[0]
                for line in lines[1:]:
                    self.content.append({'type': 'br'})
                    text_part = self._new_text_part(
                        style, text_part['ids'], part_var, text_part
                    )
                    text_part['text'] += line
                is_last_string = True
            elif isinstance(element, dict):
                self._recursive_content_parse(element, style, text_part['ids'])
                is_last_string = False
            else:
                raise TypeError(
                    'elements must be of type str or dict: {}'.format(element)
                )

        if text_part is not None and text_part['text'] == '':
            self.content.remove(text_part)

        return False



from pathlib import Path

import fontTools


def load_font(path):
    path = Path(path)
    if not path.exists():
        raise Exception("Font {} doesn't exist".format(path))

    try:
        from fontTools import ttLib
    except:
        raise ImportError(
            'You need to install library fonttools to add new fonts: pip '
            'install fonttools'
        )

    tt = ttLib.TTFont(path)
    



import copy

from .utils import (
    get_page_size, subs, parse_margin, parse_style_str, create_graphics,
    to_roman, process_style, get_paragraph_stream
)
from .standard_fonts import STANDARD_FONTS
from .base import PDFBase
from .image import PDFImage
from .text import PDFText
from .page import PDFPage
from .content import PDFContent
from .table import PDFTable

class PDF:
    def __init__(
        self, page_size='a4', portrait=True, margin=56.693,
        page_numbering_offset=0, page_numbering_style='arabic',
        font_family='Helvetica', font_size=11, font_color=0.1,
        text_align='l', line_height=1.1
    ):

        self.setup_page(page_size, portrait, margin)
        self.page_numbering_offset = page_numbering_offset
        self.page_numbering_style = page_numbering_style

        self.font_family = font_family
        self.font_size = font_size
        self.font_color = font_color
        self.text_align = text_align
        self.line_height = line_height

        self.formats = {}
        self.context = {}

        self.dests = {}
        self.uris = {}
        self.pages = []
        self.running_sections = []

        self.base = PDFBase()
        self.root = self.base.add({ 'Type': b'/Catalog'})
        self.base.trailer['Root'] = self.root.id

        self.fonts = copy.deepcopy(STANDARD_FONTS)
        self.used_fonts = {}
        self.images = {}
        self._add_font('Helvetica', 'n')

    @property
    def page(self):
        return self.pages[self._page_index]

    @property
    def page_index(self):
        return self._page_index
    @page_index.setter
    def page_index(self, page_index):
        self._page_index = page_index
        self.context['$page'] = self.get_page_number()

    @property
    def width(self):
        return self.page.width

    @property
    def height(self):
        return self.page.height

    def setup_page(self, page_size=None, portrait=None, margin=None):
        if page_size is not None:
            self.page_width, self.page_height = get_page_size(page_size)
        if portrait is not None:
            self.portrait = portrait
        if margin is not None:
            self.margin = parse_margin(margin)

    def add_page(self, page_size=None, portrait=None, margin=None):
        if page_size is not None:
            page_width, page_height = get_page_size(page_size)
        else:
            page_height, page_width = self.page_height, self.page_width

        if (portrait is None and not self.portrait) or portrait == False:
            page_height, page_width = page_width, page_height

        margin_ = copy.deepcopy(self.margin)
        if margin is not None:
            margin_.update(parse_margin(margin))

        page = PDFPage(self.base, page_width, page_height,
            **{'margin_' + side: value for side, value in margin_.items()}
        )

        self.pages.append(page)
        self.page_index = len(self.pages) - 1

        for running_section in self.running_sections:
            self._content(**running_section)
        
        page.go_to_beginning()

    def add_running_section(self, content, width, height, x, y):
        self.running_sections.append(dict(
            content=content, width=width, height=height, x=x, y=y
        ))

    def create_image(self, image):
        return PDFImage(image)

    def add_image(
        self, pdf_image, x=None, y=None, width=None, height=None, move='bottom'
    ):
        if pdf_image.image_name not in self.images:
            image_obj = self.base.add(pdf_image.pdf_obj)
            self.images[pdf_image.image_name] = image_obj.id
        else:
            image_obj = self.base[self.images[pdf_image.image_name]]

        h = pdf_image.height
        w = pdf_image.width

        if width is None and height is None:
            width = self.page.content_width
            height = width * h/w
        elif width is None:
            width = height * w/h
        elif height is None:
            height = width * h/w

        if x is not None:
            self.page.x = x
        if y is not None:
            self.page._y = y

        self.page.add_image(image_obj.id, width, height)

        if move == 'bottom':
            self.page.y += height
        if move == 'next':
            self.page.x += width

    def image(self, image, width=None, height=None, move='bottom'):
        pdf_image = self.create_image(image)
        self.add_image(pdf_image, width=width, height=height, move=move)

    def add_font(self, font_family, font_files):
        raise NotImplementedError()

    def _font_or_default(self, font_family, mode):
        return self.fonts[font_family]['n'] \
            if mode not in self.fonts[font_family] \
            else self.fonts[font_family][mode]

    def _add_font(self, font_family, mode):
        font = self._font_or_default(font_family, mode)

        font_obj = self.base.add({
            'Type': b'/Font',
            'Subtype': b'/Type1',
            'BaseFont': subs('/{}', font['base_font']),
            'Encoding': b'/WinAnsiEncoding'
        })
        self.used_fonts[(font_family, mode)] = font_obj
        return font_obj

    def _used_font(self, font_family, mode):
        font = self._font_or_default(font_family, mode)

        font_obj = self.used_fonts[(font_family, mode)] \
            if (font_family, mode) in self.used_fonts \
            else self._add_font(font_family, mode)

        self.page.add_font(font['ref'], font_obj.id)

    def _default_paragraph_style(
        self, width=None, height=None, text_align=None, line_height=None,
        indent=0
    ):
        return dict(
            width = self.page.width - self.page.margin_right - self.page.x \
                if width is None else width,
            height = self.page.height - self.page.margin_bottom - self.page.y \
                if height is None else height,
            text_align = self.text_align if text_align is None else text_align,
            line_height = self.line_height if line_height is None \
                else line_height,
            indent = indent
        )

    def _init_text(self, content):
        style = {'f': self.font_family, 's': self.font_size, 'c': self.font_color}
        if isinstance(content, str):
            content = {'style': style, '.': [content]}
        elif isinstance(content, (list, tuple)):
            content = {'style': style, '.': content}
        elif isinstance(content, dict):
            style_str = [key[1:] for key in content.keys() if key.startswith('.')]
            if len(style_str) > 0:
                style.update(parse_style_str(style_str[0], self.fonts))
            style.update(process_style(content.get('style'), self))
            content['style'] = style
        return content

    def _position_and_size(self, x=None, y=None, width=None, height=None):
        if x is not None:
            self.page.x = x
        if y is not None:
            self.page.y = y
        if width is None:
            width = self.page.width - self.page.margin_right - self.page.x
        if height is None:
            height = self.page.height - self.page.margin_bottom - self.page.y
        return self.page.x, self.page._y, width, height

    def get_page_number(self):
        page = self.page_index + 1 + self.page_numbering_offset
        return to_roman(page) if self.page_numbering_style == 'roman' else page

    def _create_text(
        self, content, width, height, text_align=None, line_height=None,
        indent=0, list_text=None, list_indent=None, list_style = None
    ):
        par_style = self._default_paragraph_style(
            width, height, text_align, line_height, indent
        )
        par_style.update({
            'list_text': list_text, 'list_indent': list_indent,
            'list_style': list_style
        })
        content = self._init_text(content)
        pdf_text = PDFText(content, fonts=self.fonts, pdf=self, **par_style)
        return pdf_text

    def _add_text(
        self, text_stream, x=None, y=None, width=None, height=None,
        graphics_stream=None, used_fonts=None, ids=None, move='bottom'
    ):
        stream = get_paragraph_stream(x, y, text_stream, graphics_stream)
        self.page.add(stream)

        for font in used_fonts:
            self._used_font(*font)

        for id_, rects in ids.items():
            if len(rects) == 0:
                continue
            if id_.startswith('$label:'):
                d = rects[0]
                x_ref = x + d[0]
                self.dests[id_[7:]] = [
                    self.page.page.id, b'/XYZ', round(x_ref, 3),
                    round(y + d[3], 3), round(x_ref/self.page.width, 3) + 1
                ]
            elif id_.startswith('$ref:'):
                for r in rects:
                    self.page.add_reference(
                        id_[5:],
                        [
                            round(x + r[0], 3), round(y + r[1], 3),
                            round(x + r[2], 3), round(y + r[3], 3)
                        ]
                    )
            elif id_.startswith('$uri:'):
                link = id_[5:]
                if not link in self.uris:
                    uri = self.base.add(
                        {'Type': b'/Action', 'S': b'/URI', 'URI': link}
                    )
                    self.uris[link] = uri.id

                for r in rects:
                    self.page.add_link(
                        self.uris[link],
                        [
                            round(x + r[0], 3), round(y + r[1], 3),
                            round(x + r[2], 3), round(y + r[3], 3)
                        ]
                    )

        if move == 'bottom':
            self.page.y += height
        if move == 'next':
            self.page.x += width

    def _text(
        self, content, x=None, y=None, width=None, height=None, text_align=None,
        line_height=None, indent=0, list_text=None, list_indent=None,
        list_style = None, move='bottom'
    ):
        x, y, width, height = self._position_and_size(x, y, width, height)
        if isinstance(content, PDFText):
            pdf_text = content
            pdf_text.run(x, y, width, height)
        else:
            pdf_text = self._create_text(
                content, width, height, text_align, line_height, indent,
                list_text, list_indent, list_style
            )
            pdf_text.run()
            pdf_text.move(x, y)

        self._add_text(move=move, **pdf_text.result)
        return pdf_text

    def text(
        self, content, text_align=None, line_height=None, indent=0,
        list_text=None, list_indent=None, list_style=None
    ):
        pdf_text = self._text(
            content, x=self.page.margin_left, width=self.page.content_width,
            text_align=text_align, line_height=line_height, indent=indent,
            list_text=list_text, list_indent=list_indent, list_style=list_style
        )
        while not pdf_text.finished:
            self.add_page()
            pdf_text = self._text(pdf_text, self.page.content_width,
                self.page.content_height, self.page.margin_left,
                self.page.margin_top
            )

    def _default_content_style(self):
        return dict(
            f=self.font_family, s=self.font_size, c=self.font_color,
            text_align=self.text_align, line_height=self.line_height, indent=0
        )

    def _create_table(
        self, content, width, height, x=None, y=None, widths=None, style=None,
        borders=None, fills=None
    ):
        style_ = self._default_content_style()
        style_.update(process_style(style, self))
        pdf_table = PDFTable(
            content, self.fonts, width, height, x, y,
            widths, style_, borders, fills, self
        )
        return pdf_table

    def _table(
        self, content, width=None, height=None, x=None, y=None, widths=None,
        style=None, borders=None, fills=None, move='bottom'
    ):
        x, y, width, height = self._position_and_size(x, y, width, height)

        if isinstance(content, PDFTable):
            pdf_table = content
            pdf_table.run(x, y, width, height)
        else:
            pdf_table = self._create_table(
                content, width, height, x, y, widths, style, borders, fills
            )
            pdf_table.run()

        self._add_graphics([*pdf_table.fills, *pdf_table.lines])
        self._add_parts(pdf_table.parts_)

        if move == 'bottom':
            self.page.y += pdf_table.current_height
        if move == 'next':
            self.page.x += width

        return pdf_table

    def table(
        self, content, widths=None, style=None, borders=None, fills=None
    ):
        pdf_table = self._table(
            content, widths=widths, style=style, borders=borders, fills=fills,
            x=self.page.margin_left, width=self.page.content_width
        )
        while not pdf_table.finished:
            self.add_page()
            pdf_table = self._table(
                pdf_table, self.page.content_width, self.page.content_height,
                self.page.margin_left, self.page.margin_top
            )

    def _create_content(self, content, width=None, height=None, x=None, y=None):
        style = self._default_content_style()
        content = content.copy()
        style.update(process_style(content.get('style'), self))
        content['style'] = style
        pdf_content = PDFContent(content, self.fonts, width, height, x, y, self)
        return pdf_content

    def _content(
        self, content, width=None, height=None, x=None, y=None, move='bottom'
    ):
        x, y, width, height = self._position_and_size(x, y, width, height)

        if isinstance(content, PDFContent):
            pdf_content = content
            pdf_content.run(x, y, width, height)
        else:
            pdf_content = self._create_content(content, width, height, x, y)
            pdf_content.run()

        self._add_graphics([*pdf_content.fills,*pdf_content.lines])
        self._add_parts(pdf_content.parts_)

        if move == 'bottom':
            self.page.y += pdf_content.current_height
        if move == 'next':
            self.page.x += width

        return pdf_content

    def content(self, content):
        pdf_content = self._content(
            content, x=self.page.margin_left, width=self.page.content_width
        )
        while not pdf_content.finished:
            self.add_page()
            pdf_content = self._content(pdf_content,
                self.page.content_width, self.page.content_height,
                self.page.margin_left, self.page.margin_top
            )

    def _add_graphics(self, graphics):
        stream = create_graphics(graphics)
        self.page.add(stream)

    def _add_parts(self, parts):
        for part in parts:
            part = part.copy()
            type_ = part.pop('type')
            if type_ == 'paragraph':
                self._add_text(**part)
            elif type_ == 'image':
                self.add_image(**part)

    def _build_pages_tree(self, page_list, first_level = True):
        new_page_list = []
        count = 0
        for page in page_list:
            if first_level:
                page_size = [page.width, page.height]
                page = page.page
                page['MediaBox'] = [0, 0] + page_size

            if count % 6 == 0:
                new_page_list.append(
                    self.base.add({'Type': b'/Pages', 'Kids': [], 'Count': 0})
                )
                count += 1

            last_parent = new_page_list[-1]
            page['Parent'] = last_parent.id
            last_parent['Kids'].append(page.id)
            last_parent['Count'] += 1

        if count == 1:
            self.root['Pages'] = new_page_list[0].id
        else:
            self._build_pages_tree(new_page_list, False)

    def _build_dests_tree(self, keys, vals, first_level=True):
        k = 7
        new_keys = []
        new_vals = []
        i = 0
        length = len(keys)
        obj = None
        count = 0

        while i < length:
            key = keys[i]
            val = vals[i]
            if i % k == 0:
                count += 1
                obj = self.base.add({})
                obj['Limits'] = [key if first_level else val[0], None]
                if first_level: obj['Names'] = []
                else: obj['Kids'] = []

            if first_level:
                obj['Names'].append(key)
                obj['Names'].append(val)
            else:
                obj['Kids'].append(key)

            if (i + 1) % k == 0 or (i + 1) == length:
                obj['Limits'][1] = key if first_level else val[1]
                new_keys.append(obj.id)
                new_vals.append(obj['Limits'])

            i += 1

        if count == 1:
            del obj['Limits']
            if not 'Names' in self.root: self.root['Names'] = {}
            self.root['Names']['Dests'] = obj.id
        else:
            self._build_dests_tree(new_keys, new_vals, False)

    def _build_dests(self):
        dests = list(self.dests.keys())
        if len(dests) == 0:
            return
        dests.sort()
        self._build_dests_tree(dests, [self.dests[k] for k in dests])

    def output(self, buffer):
        if len(self.pages) == 0:
            raise Exception("pdf doesn't have any pages")
        self._build_pages_tree(self.pages)
        self._build_dests()
        self.base.output(buffer)



from copy import deepcopy

from .color import PDFColor
from .text import PDFText
from .image import PDFImage
from .utils import parse_style_str, process_style


PARAGRAPH_PROPS = (
    'text_align', 'line_height', 'indent', 'list_text',
    'list_style', 'list_indent'
)

TABLE_PROPS = ('widths', 'borders', 'fills')

class PDFTable:
    def __init__(
        self, content, fonts, width, height, x=0, y=0, widths=None,
        style=None, borders=None, fills=None, pdf=None
    ):
        if not isinstance(content, (list, tuple)):
            raise Exception('content must be a list or tuple')

        self.setup(x, y, width, height)
        self.content = content
        self.current_index = 0
        self.fonts = fonts
        self.pdf = pdf
        self.finished = False
        if len(content) == 0 or len(content[0]) == 0:
            return

        self.cols_count = len(content[0])
        self.delayed = {}

        if widths is not None:
            if not isinstance(widths, (list, tuple)):
                raise Exception('widths must be a list or tuple')
            if len(widths) != self.cols_count:
                raise Exception('widths count must be equal to cols count')
            try:
                widths_sum = sum(widths)
                self.widths = [w/widths_sum for w in widths]
            except TypeError:
                raise Exception('widths must be numbers')
            except ZeroDivisionError:
                raise Exception('sum of widths must be greater than zero')
        else:
            self.widths = [1 / self.cols_count] * self.cols_count

        self.style = {'cell_margin': 5, 'cell_fill': None}
        self.style.update(process_style(style, self.pdf))
        self.set_default_border()
        self.setup_borders([] if borders is None else borders)
        self.setup_fills([] if fills is None else fills)
        self.first_row = True

    def setup(self, x=None, y=None, width=None, height=None):
        if x is not None:
            self.x = x
        if y is not None:
            self.y = y
        if width is not None:
            self.width = width
        if height is not None:
            self.height = height

    def set_default_border(self):
        self.default_border = {}
        self.default_border['width'] = self.style.pop('border_width', 0.5)
        color = self.style.pop('border_color', 0)
        self.default_border['color'] = PDFColor(color, True)
        self.default_border['style'] = self.style.pop('border_style', 'solid')

    def parse_style_string(self, pos, counts):
        range_strs = pos.split(';')
        if len(range_strs) == 1:
            range_strs.append(':')
        ranges = [
            self.parse_range_string(range_str, count)
            for range_str, count in zip(range_strs, counts)
        ]
        for i in ranges[0]:
            for j in ranges[1]:
                yield i, j

    def range_generator(self, list_str, count):
        for i in list_str.split(','):
            num = int(i)
            yield num if num >= 0 else count + num

    def parse_range_string(self, data, count):
        data = ':' if data == '' else data
        if ':' in data:
            parts = data.split(':')
            num = 0 if parts[0].strip() == '' else int(parts[0])
            parts[0] = num if num >= 0 else count + num
            if len(parts) > 1:
                num = count if parts[1].strip() == '' else int(parts[1])
                parts[1] = num if num >= 0 else count + num
            if len(parts) > 2:
                parts[2] = 1 if parts[2].strip() == '' else int(parts[2])
            return range(*parts)
        else:
            return [i for i in self.range_generator(data, count)]

    def setup_borders(self, borders):
        rows = len(self.content)
        cols = self.cols_count
        self.borders_h = {}
        self.borders_v = {}
        for b in borders:
            border = deepcopy(deepcopy(self.default_border))
            border.update(b)
            border['color'] = PDFColor(border['color'], True)
            pos = border.pop('pos', None)
            if pos is None:
                continue
            is_vert = pos[0].lower() == 'v'
            counts = (rows, cols + 1) if is_vert else (rows + 1, cols)
            border_l = self.borders_v if is_vert else self.borders_h
            for i, j in self.parse_style_string(pos[1:], counts):
                first = border_l.setdefault(i, {})
                first[j] = border

    def get_border(self, i, j, is_vert):
        border_l = self.borders_v if is_vert else self.borders_h
        if i in border_l and j in border_l[i]:
            return deepcopy(border_l[i][j])
        else:
            return deepcopy(self.default_border)

    def setup_fills(self,  fills):
        v_count = len(self.content)
        h_count = self.cols_count
        self.fills_defs = {}
        for f in fills:
            fill = f['color']
            pos = f.get('pos', None)
            if pos is None:
                continue
            for i, j in self.parse_style_string(pos, (v_count, h_count)):
                first = self.fills_defs.setdefault(i, {})
                first[j] = fill

    def compare_borders(self, a, b):
        return all(a[attr] == b[attr] for attr in ['width', 'color', 'style'])

    def process_borders(self, col, border_left, border_top):
        vert_correction = border_top.get('width', 0)/2
        if not self.top_lines_interrupted:
            aux = self.top_lines[-1].get('width', 0)/2
            if aux > vert_correction:
                vert_correction = aux

        v_line = self.vert_lines[col]
        horiz_correction = border_left.get('width', 0)/2
        if not v_line['interrupted']:
            v_line['list'][-1]['y2'] = self.y_abs - vert_correction
            aux = v_line['list'][-1].get('width', 0)/2
            if aux > horiz_correction:
                horiz_correction = aux

        if not self.top_lines_interrupted:
            self.top_lines[-1]['x2'] += horiz_correction

        if border_top.get('width', 0) > 0:
            x2 = self.x_abs + self.widths[col] * self.width
            if (
                not self.top_lines_interrupted and
                self.compare_borders(self.top_lines[-1], border_top)
            ):
                self.top_lines[-1]['x2'] = x2
            else:
                border_top.update(dict(
                    type='line', y1=self.y_abs,
                    y2=self.y_abs, x2=x2, x1=self.x_abs - horiz_correction
                ))
                self.top_lines.append(border_top)
                self.top_lines_interrupted = False
        else:
            self.top_lines_interrupted = True

        if border_left.get('width', 0) > 0:
            if not (
                not v_line['interrupted'] and
                self.compare_borders(v_line['list'][-1], border_left)
            ):
                border_left.update(dict(
                    type='line', x1=self.x_abs,
                    x2=self.x_abs, y1=self.y_abs + vert_correction
                ))
                v_line['list'].append(border_left)
                v_line['interrupted'] = False
        else:
            v_line['interrupted'] = True

    def get_delayed_cols_n(self):
        return sum(v['colspan'] for v in self.delayed.values())

    def can_continue(self):
        if self.get_delayed_cols_n() == self.cols_count:
            return False
        for col in self.delayed:
            if not col in self.rowspan:
                return False
        return True

    def run(self, x=None, y=None, width=None, height=None):
        self.setup(x, y, width, height)
        self.parts_ = []
        self.lines = []
        self.fills = []
        self.fills_mem = {}
        self.heights_mem = {}
        self.rowspan = {}
        self.vert_lines = [
            {'list': [], 'interrupted': True}
            for i in range(self.cols_count + 1)
        ]
        self.current_height = 0

        can_continue = True
        if len(self.delayed) > 0:
            row = [self.delayed.get(i) for i in range(self.cols_count)]
            action = self.add_row(row, True)
            if action == 'continue':
                self.current_index += 1

        cancel = False
        if can_continue:
            while self.current_index < len(self.content):
                action = self.add_row(self.content[self.current_index])
                if action == 'cancel':
                    cancel = True
                    break
                elif action == 'continue':
                    self.current_index += 1
                else:
                    break

        if cancel:
            self.parts_ = []
            self.lines = []
            self.fills = []
            return

        self.top_lines = []
        self.top_lines_interrupted = True
        self.y_abs = self.y - self.current_height
        for col in range(self.cols_count):
            self.x_abs = self.x + sum(self.widths[0:col]) * self.width
            border_top = self.get_border(self.current_index, col, False)
            self.process_borders(col, {}, border_top)

        self.x_abs = self.x + self.width
        self.process_borders(self.cols_count, {}, {})
        self.lines.extend(self.top_lines)
        self.lines.extend(
            line for vert_line in self.vert_lines for line in vert_line['list']
        )

        if self.current_index >= len(self.content) and len(self.delayed) == 0:
            self.finished = True

    def add_row(self, row, is_delayed=False):
        if not len(row) == self.cols_count:
            error = 'row {} should be of length {}.'
            raise Exception(error.format(self.current_index+1, self.cols_count))

        self.max_height = 0
        self.row_max_height = self.height - self.current_height
        self.colspan = 0
        self.top_lines = []
        self.top_lines_interrupted = True
        self.is_rowspan = False
        self.y_abs = self.y - self.current_height
        for col, element in enumerate(row):
            move_next = self.add_cell(col, element, is_delayed)
            if move_next:
                continue

        if self.first_row:
            self.first_row = False
            if self.max_height == 0:
                return 'cancel'

        ret = self.can_continue()

        self.x_abs = self.x + self.width
        last_border = self.get_border(self.current_index, len(row), True)
        self.process_borders(len(row), last_border, {})

        self.lines.extend(self.top_lines)

        for col in self.heights_mem:
            self.heights_mem[col] -= self.max_height

        for col, fill in list(self.fills_mem.items()):
            fill['height'] += self.max_height
            if not fill.get('add_later', False) or not ret:
                fill['y'] -= fill['height']
                self.fills.append(fill)
                self.fills_mem.pop(col)

        self.current_height += self.max_height

        return 'continue' if ret else 'interrupt'
        
    def get_cell_dimensions(
        self, col, border_left, border_top, cell_style, rowspan, colspan
    ):
        border_right = self.get_border(
            self.current_index, col + colspan + 1, True
        )
        border_bottom = self.get_border(
            self.current_index + rowspan + 1, col, False
        )

        full_width = sum(self.widths[col:col + colspan + 1]) * self.width
        padd_x_left = border_left.get('width', 0) / 2 + \
            cell_style['cell_margin_left']
        padd_x_right = border_right.get('width', 0) / 2 + \
            cell_style['cell_margin_right']
        padd_x = padd_x_left + padd_x_right
        padd_y_top = border_top.get('width', 0) / 2 + \
            cell_style['cell_margin_top']
        padd_y_bottom = border_bottom.get('width', 0) / 2 + \
            cell_style['cell_margin_bottom']
        padd_y = padd_y_top + padd_y_bottom
        x = self.x_abs + padd_x_left
        y = self.y_abs - padd_y_top
        width = full_width - padd_x
        height = self.row_max_height - padd_y
        return x, y, width, height, padd_x, padd_y

    def is_span(self, col, border_left, border_top, is_delayed):
        rowspan_memory = self.rowspan.get(col, None)
        can_continue = False
        if rowspan_memory is None:
            if self.colspan > 0:
                self.colspan -= 1
                border_left['width'] = 0
                if self.is_rowspan:
                    border_top['width'] = 0
                    if self.colspan == 0:
                        self.is_rowspan = False
                can_continue = True
        elif not is_delayed:
            rowspan_memory['rows'] -= 1
            self.colspan = rowspan_memory['cols']
            self.is_rowspan = True
            border_top['width'] = 0

            if rowspan_memory['rows'] == 0:
                self.rowspan.pop(col)
                if col in self.fills_mem:
                    self.fills_mem[col]['add_later'] = False
                if self.heights_mem.get(col, 0) > self.max_height:
                    self.max_height = self.heights_mem[col]
            can_continue = True
        return can_continue

    def get_cell_style(self, element):
        if isinstance(element, dict) and 'delayed' in element:
            cell_style = element['cell_style']
            style = {}
        else:
            style = deepcopy(self.style)

            if element is None:
                element = ''
            if not isinstance(element, (dict, str, list, tuple)):
                element = str(element)
            if isinstance(element, (str, list, tuple)):
                element = {'.': element}

            if not isinstance(element, dict):
                raise TypeError(
                    'Elements must be of type dict, str, list or tuple:'
                    + str(element)
                )

            keys = [key for key in element.keys() if key.startswith('.')]
            if len(keys) > 0:
                style.update(parse_style_str(keys[0][1:], self.fonts))
            style.update(process_style(element.get('style'), self.pdf))
            cell_style = {}
            attr = 'cell_margin'
            for side in ('top', 'right', 'bottom', 'left'):
                cell_style[attr + '_' + side] = style.pop(
                    attr + '_' + side, style.get(attr, None)
                )
            style.pop(attr, None)
            cell_style['cell_fill'] = style.pop('cell_fill', None)
        element = deepcopy(element)
        return element, style, cell_style

    def setup_cell_fill(self, col, cell_style, width, rowspan):
        fill_color = cell_style.get('cell_fill',
            self.fills_defs[(self.current_index, col)]
            if (self.current_index, col) in self.fills_defs
            else None
        )

        if fill_color is not None:
            self.fills_mem[col] = { 'type': 'fill', 'x': self.x_abs,
                'y': self.y_abs, 'width': width, 'height': 0,
                'color': PDFColor(fill_color, stroke=False)
            }
            if rowspan > 0:
                self.fills_mem[col]['add_later'] = True

    def is_delayed_type(self, el, type_):
        return 'delayed' in el and el['type'] == type_

    def is_type(self, el, type_):
        return type_ in el or self.is_delayed_type(el, type_)

    def add_cell(self, col, element, is_delayed):
        self.x_abs = self.x + sum(self.widths[0:col]) * self.width
        border_left = self.get_border(self.current_index, col, True)
        border_top = self.get_border(self.current_index, col, False)
        can_continue = self.is_span(col, border_left, border_top, is_delayed)
        self.process_borders(col, border_left, border_top)
        if can_continue:
            return True

        element, style, cell_style = self.get_cell_style(element)

        colspan_original = element.get('colspan', 1)
        self.colspan = colspan_original - 1
        row_added = element.get('row', self.current_index)
        rowspan = element.get('rowspan', 1) - 1 - self.current_index + \
            row_added
        if rowspan > 0:
            self.rowspan[col] = {'rows': rowspan, 'cols': self.colspan}

        x, y, width, height, padd_x, padd_y = self.get_cell_dimensions(
            col, border_left, border_top, cell_style, rowspan, self.colspan
        )

        delayed = {
            'cell_style': cell_style, 'colspan': colspan_original,
            'rowspan': element.get('rowspan', 1), 'row': row_added
        }

        self.setup_cell_fill(col, cell_style, width + padd_x, rowspan)

        keys = [key for key in element.keys() if key.startswith('.')]

        real_height = 0
        if len(keys) > 0 or self.is_delayed_type(element, 'text'):
            real_height = self.process_text(
                col, element, x, y, width, height, style, keys, delayed
            )
        elif self.is_type(element, 'image'):
            real_height = self.process_image(
                col, element, x, y, width, height, delayed
            )
        elif self.is_type(element, 'content'):
            real_height = self.process_content(
                col, element, x, y, width, height, style, delayed
            )

        elif self.is_type(element, 'table'):
            real_height = self.process_table(
                col, element, x, y, width, height, style, delayed
            )

        real_height += padd_y if real_height>0 else (0 if self.first_row else 4)
        if rowspan > 0:
            self.heights_mem[col] = real_height
            real_height = 0
        if real_height > self.max_height:
            self.max_height = real_height

        return False

    def process_text(
        self, col, element, x, y, width, height, style, keys, delayed
    ):
        if 'delayed' in element:
            pdf_text = element['delayed']
            pdf_text.setup(x, y, width, height)
        else:
            par_style = {
                v: style.get(v) for v in PARAGRAPH_PROPS if v in style
            }
            key = keys[0]
            pdf_text = PDFText(
                {key: element[key], 'style': style},
                width, height, x, y,
                fonts=self.fonts, pdf=self.pdf, **par_style
            )

        result = pdf_text.run()
        result['type'] = 'paragraph'
        self.parts_.append(result)

        if not pdf_text.finished:
            delayed.update({'delayed': pdf_text, 'type': 'text'})
            self.delayed[col] = delayed
        else:
            self.delayed.pop(col, None)
        return pdf_text.current_height
    
    def process_image(self, col, element, x, y, width, height, delayed):
        if 'delayed' in element:
            pdf_image = element['delayed']
        else:
            pdf_image = PDFImage(element['image'])
        img_width = width
        img_height = img_width * pdf_image.height / pdf_image.width

        real_height = 0
        if img_height < height:
            real_height = img_height
            self.parts_.append({
                'type': 'image', 'pdf_image': pdf_image,
                'x': x, 'y': y - img_height,
                'width': img_width, 'height': img_height
            })
            self.delayed.pop(col, None)
        else:
            delayed.update({'delayed': pdf_image, 'type': 'image'})
            self.delayed[col] = delayed
        return real_height
    
    def process_content(
        self, col, element, x, y, width, height, style, delayed
    ):
        if 'delayed' in element and element['type'] == 'content':
            pdf_content = element['delayed']
            pdf_content.setup(x, y, width, height)
        else:
            element['style'] = style
            pdf_content = PDFContent(
                element, self.fonts, width, height, x, y, self.pdf
            )

        pdf_content.run()

        self.parts_.extend(pdf_content.parts_)
        self.lines.extend(pdf_content.lines)
        self.fills.extend(pdf_content.fills)

        if not pdf_content.finished:
            delayed.update({'delayed': pdf_content, 'type': 'content'})
            self.delayed[col] = delayed
        else:
            self.delayed.pop(col, None)
        
        return pdf_content.current_height

    def process_table(self, col, element, x, y, width, height, style, delayed):
        if 'delayed' in element and element['type'] == 'table':
            pdf_table = element['delayed']
            pdf_table.setup(x, y, width, height)
        else:
            table_props = {
                v: element.get(v) for v in TABLE_PROPS if v in element
            }
            pdf_table = PDFTable(
                element['table'], self.fonts, width, height, x, y,
                style=style, pdf=self.pdf, **table_props
            )

        pdf_table.run()

        self.parts_.extend(pdf_table.parts_)
        self.lines.extend(pdf_table.lines)
        self.fills.extend(pdf_table.fills)

        if not pdf_table.finished:
            delayed.update({'delayed': pdf_table, 'type': 'table'})
            self.delayed[col] = delayed
        else:
            self.delayed.pop(col, None)

        return pdf_table.current_height

from .content import PDFContent



import zlib

def encode_stream(stream, filter, parameters = {}):

	if filter == b'/ASCIIHexDecode':
		raise NotImplementedError('/ASCIIHexDecode')
	elif filter == b'/ASCII85Decode':
		raise NotImplementedError('/ASCII85Decode')
	elif filter == b'/LZWDecode':
		raise NotImplementedError('/LZWDecode')
	elif filter == b'/FlateDecode':
		return flate_encode(stream)
	elif filter == b'/RunLengthDecode':
		raise NotImplementedError('/RunLengthDecode')
	elif filter == b'/CCITTFaxDecode':
		raise NotImplementedError('/CCITTFaxDecode')
	elif filter == b'/JBIG2Decode':
		raise NotImplementedError('/JBIG2Decode')
	elif filter == b'/DCTDecode':
		raise NotImplementedError('/DCTDecode')
	elif filter == b'/JPXDecode':
		raise NotImplementedError('/JPXDecode')
	elif filter == b'/Crypt':
		raise NotImplementedError('/Crypt')

def flate_encode(stream):
    return zlib.compress(stream)




from copy import deepcopy
from .pdf import PDF

STYLE_PROPS = dict(
    f='font_family', s='font_size', c='font_color',
    text_align='text_align', line_height='line_height'
)

PAGE_PROPS = ('page_size', 'portrait', 'margin')

class PDFDocument:
    def __init__(self, document, context=None):
        context = {} if context is None else context
        style = deepcopy(document.get('style', {}))
        style_args = {
            v: style.pop(k) for k, v in STYLE_PROPS.items() if k in style
        }
        page_style = document.get('page_style', {})
        page_args = {p: page_style[p] for p in PAGE_PROPS if p in page_style}

        self.pdf = PDF(**page_args, **style_args)
        self.pdf.formats = {}
        self.pdf.formats.setdefault('$footnote', {'r': 0.5, 's': 6})
        self.pdf.formats.setdefault('$footnotes', {'s': 10, 'c': 0})
        self.pdf.formats.update(document.get('formats', {}))
        self.pdf.context.update(context)

        self.defs = document.get('defs', {})
        self.style = style

        self.sections = document.get('sections', [])

        self.footnotes = []
        self.traverse_document_footnotes(self.sections)

        self.footnotes_margin = 10

    def traverse_document_footnotes(self, element):
        if isinstance(element, (list, tuple)):
            for child in element:
                self.traverse_document_footnotes(child)
        elif isinstance(element, dict):
            if 'footnote' in element:
                element.setdefault('ids', [])
                name = '$footnote:' + str(len(self.footnotes))
                element['ids'].append(name)
                element['style'] = '$footnote'
                element['var'] = name
                self.pdf.context[name] = '0'

                footnote = element['footnote']

                if not isinstance(footnote, (dict, str, list, tuple)):
                    footnote = str(footnote)
                if isinstance(footnote, (str, list, tuple)):
                    footnote = {'.': footnote}

                if not isinstance(footnote, dict):
                    raise TypeError(
                        'footnotes must be of type dict, str, list or tuple:{}'
                        .format(footnote)
                    )

                self.footnotes.append(footnote)
            else:
                for value in element.values():
                    if isinstance(value, (list, tuple, dict)):
                        self.traverse_document_footnotes(value)

    def set_running_sections(self, running_sections):
        self.pdf.running_sections = []
        for name in running_sections:
            section = deepcopy(self.defs[name])
            margin = self.pdf.margin
            if section.get('width') in ['left', 'right']:
                section['width'] = margin[section.get('width')]
            if section.get('width') == 'full':
                section['width'] = self.pdf.page_width
            if section.get('height') in ['top', 'bottom']:
                section['height'] = margin[section.get('height')]
            if section.get('height') == 'full':
                section['height'] = self.pdf.page_height
            if section.get('x') == 'left':
                section['x'] = margin['left']
            if section.get('x') == 'right':
                section['x'] = self.pdf.page_width - margin['right']
            if section.get('y') == 'top':
                section['y'] = margin['top']
            if section.get('y') == 'bottom':
                section['y'] = self.pdf.page_height - margin['bottom']

            width = section.get('width', (
                self.pdf.page_width - margin['right'] - margin['left']
            ))
            height = section.get('height', (
                self.pdf.page_height - margin['top'] - margin['bottom']
            ))
            x = section.get('x', 0)
            y = section.get('y', 0)
            self.pdf.add_running_section(section, width, height, x, y)

    def run(self):
        for section in self.sections:
            self.process_section(section)

    def process_section(self, section):
        page_style = section.get('page_style', {})
        page_args = {p: page_style[p] for p in PAGE_PROPS if p in page_style}
        self.pdf.setup_page(**page_args)
        running_sections = section.get('running_sections', [])
        self.set_running_sections(running_sections)

        pdf = self.pdf
        self.width = pdf.page_width - pdf.margin['right'] - pdf.margin['left']
        self.y = pdf.page_height - pdf.margin['top']
        self.height = self.y - pdf.margin['bottom']
        self.x = pdf.margin['left']

        if 'page_numbering_offset' in page_style:
            self.pdf.page_numbering_offset = page_style['page_numbering_offset']
        if 'page_numbering_style' in page_style:
            self.pdf.page_numbering_style = page_style['page_numbering_style']
        if page_style.get('page_numbering_reset', False):
            self.pdf.page_numbering_offset = -len(self.pdf.pages)

        self.section = self.pdf._create_content(
            section, self.width, self.height, self.x, self.y
        )

        self.add_pages()

    def add_pages(self):
        while not self.section.finished:
            self.add_page()

    def add_page(self):
        self.pdf.add_page()
        content_part = self.section.pdf_content_part
        first_page = content_part is None
        if first_page:
            section_element_index = 0
            section_delayed = []
            children_indexes = []
        else:
            section_element_index = deepcopy(content_part.section_element_index)
            section_delayed = deepcopy(content_part.section_delayed)
            children_indexes = deepcopy(content_part.children_indexes)
        
        self.section.run(height=self.height)
        if first_page:
            content_part = self.section.pdf_content_part
        footnotes_obj = self.process_footnotes()

        if footnotes_obj is None:
            self.pdf._add_graphics([*self.section.fills,*self.section.lines])
            self.pdf._add_parts(self.section.parts_)
            self.pdf.page._y -= self.section.current_height
        else:
            footnotes_height = footnotes_obj.current_height
            if footnotes_height >= self.height - self.footnotes_margin - 20:
                raise Exception(
                    "footnotes are very large and don't fit in one page"
                )
            new_height = self.height - footnotes_obj.current_height \
                - self.footnotes_margin
            
            content_part.section_element_index = section_element_index
            content_part.section_delayed = section_delayed
            content_part.children_indexes = children_indexes

            self.pdf._content(self.section, height=new_height)

            footnotes_obj = self.process_footnotes()
            
            if footnotes_obj is not None:
                self.pdf.page._y = self.pdf.margin['bottom'] + footnotes_height

                x_line = round(self.pdf.page.x, 3)
                y_line = round(self.pdf.page._y + self.footnotes_margin/2, 3)
                self.pdf.page.add(' q 0 G 0.5 w {} {} m {} {} l S Q'.format(
                    x_line, y_line, x_line + 150, y_line
                ))
                self.pdf._content(footnotes_obj, height=self.height)

    def check_footnote(self, ids, page_footnotes):
        for id_, rects in ids.items():
            if len(rects) == 0:
                continue
            if id_.startswith('$footnote:'):
                index = int(id_[10:])
                page_footnotes.append(self.footnotes[index])
                self.pdf.context[id_] = len(page_footnotes)

    def check_footnotes(self, page_footnotes):
        for part in self.section.parts_:
            if part['type'] == 'paragraph':
                self.check_footnote(part['ids'], page_footnotes)
        
    def get_footnotes_obj(self, page_footnotes):
        content = {'style': '$footnotes', 'content': []}
        for index, footnote in enumerate(page_footnotes):
            footnote = deepcopy(footnote)
            style = footnote.setdefault('style', {})
            style.update(dict(list_text=index + 1, list_style='$footnote'))
            content['content'].append(footnote)

        footnote_obj = self.pdf._create_content(
            content, self.width, self.height, self.x, self.y
        )
        footnote_obj.run()
        return footnote_obj
    
    def process_footnotes(self):
        page_footnotes = []
        self.check_footnotes(page_footnotes)
        if len(page_footnotes) == 0:
            return None
        return self.get_footnotes_obj(page_footnotes)

    def output(self, buffer):
        self.pdf.output(buffer)

def build_pdf(document, buffer, context=None):
    doc = PDFDocument(document, context)
    doc.run()
    doc.output(buffer)